package workers

import (
	"testing"

	luajit "git.sharkk.net/Sky/LuaJIT-to-Go"
)

// TestWorkerWithGoFunctions tests registering and calling Go functions from Lua
func TestWorkerWithGoFunctions(t *testing.T) {
	// Create config with Go functions
	config := NewWorkerConfig()
	config.AddFunction("add", func(s *luajit.State) int {
		a := s.ToNumber(1)
		b := s.ToNumber(2)
		s.PushNumber(a + b)
		return 1
	})
	config.AddFunction("concat", func(s *luajit.State) int {
		a := s.ToString(1)
		b := s.ToString(2)
		s.PushString(a + b)
		return 1
	})

	// Create pool with config
	pool, err := NewPool(2, config)
	if err != nil {
		t.Fatalf("Failed to create pool: %v", err)
	}
	defer pool.Shutdown()

	// Test using the Go functions from Lua
	bytecode := createTestBytecode(t, `
		local sum = add(5, 7)
		local str = concat("hello", "world")
		return {sum = sum, str = str}
	`)

	result, err := pool.Submit(bytecode, nil)
	if err != nil {
		t.Fatalf("Failed to submit job: %v", err)
	}

	// Check results
	resultMap, ok := result.(map[string]any)
	if !ok {
		t.Fatalf("Expected map result, got %T", result)
	}

	if resultMap["sum"] != 12.0 {
		t.Errorf("Expected sum=12, got %v", resultMap["sum"])
	}

	if resultMap["str"] != "helloworld" {
		t.Errorf("Expected str=helloworld, got %v", resultMap["str"])
	}
}

// TestWorkerWithModules tests loading and using Lua modules
func TestWorkerWithModules(t *testing.T) {
	// Create config with modules
	config := NewWorkerConfig()

	// Add a module as a string
	mathModule := `
		local math_utils = {}
		
		function math_utils.add(a, b)
			return a + b
		end
		
		function math_utils.multiply(a, b)
			return a * b
		end
		
		return math_utils
	`
	config.AddModule("math_utils", mathModule)

	// Add another module
	stringModule := `
		local string_utils = {}
		
		function string_utils.concat(a, b)
			return a .. b
		end
		
		function string_utils.reverse(s)
			return string.reverse(s)
		end
		
		return string_utils
	`
	config.AddModule("string_utils", stringModule)

	// Create pool with config
	pool, err := NewPool(2, config)
	if err != nil {
		t.Fatalf("Failed to create pool: %v", err)
	}
	defer pool.Shutdown()

	// Test using the modules
	bytecode := createTestBytecode(t, `
		local math = require("math_utils")
		local str = require("string_utils")
		
		local sum = math.add(5, 7)
		local product = math.multiply(3, 4)
		local combined = str.concat("hello", "world")
		local reversed = str.reverse("abcdef")
		
		return {
			sum = sum,
			product = product,
			combined = combined,
			reversed = reversed
		}
	`)

	result, err := pool.Submit(bytecode, nil)
	if err != nil {
		t.Fatalf("Failed to submit job: %v", err)
	}

	// Check results
	resultMap, ok := result.(map[string]any)
	if !ok {
		t.Fatalf("Expected map result, got %T", result)
	}

	if resultMap["sum"] != 12.0 {
		t.Errorf("Expected sum=12, got %v", resultMap["sum"])
	}

	if resultMap["product"] != 12.0 {
		t.Errorf("Expected product=12, got %v", resultMap["product"])
	}

	if resultMap["combined"] != "helloworld" {
		t.Errorf("Expected combined=helloworld, got %v", resultMap["combined"])
	}

	if resultMap["reversed"] != "fedcba" {
		t.Errorf("Expected reversed=fedcba, got %v", resultMap["reversed"])
	}
}

// TestCustomInitFunction tests using a custom initialization function
func TestCustomInitFunction(t *testing.T) {
	// Create config with custom init function
	config := NewWorkerConfig()

	// Set custom init function that sets a global variable
	config.SetCustomInit(func(s *luajit.State) error {
		return s.DoString(`
			-- Create a helper utility in the global environment
			-- We're adding this through custom init rather than as a module
			function split_string(str, sep)
				local result = {}
				for part in string.gmatch(str .. sep, "(.-)" .. sep) do
					table.insert(result, part)
				end
				return result
			end
		`)
	})

	// Create pool with config
	pool, err := NewPool(2, config)
	if err != nil {
		t.Fatalf("Failed to create pool: %v", err)
	}
	defer pool.Shutdown()

	// Test using the custom initialized function
	bytecode := createTestBytecode(t, `
		local parts = split_string("a,b,c,d", ",")
		return parts
	`)

	result, err := pool.Submit(bytecode, nil)
	if err != nil {
		t.Fatalf("Failed to submit job: %v", err)
	}

	// Result should be an array
	resultMap, ok := result.(map[string]any)
	if !ok {
		t.Fatalf("Expected map result, got %T", result)
	}

	// Check if the empty key exists for array values
	if _, hasArray := resultMap[""]; !hasArray {
		t.Fatalf("Expected array to be present at empty key, got: %v", resultMap)
	}

	// Check array elements
	array, ok := resultMap[""].([]float64)
	if !ok {
		t.Fatalf("Expected []float64 under empty key, got %T", resultMap[""])
	}

	// Check array length (should have 4 elements)
	if len(array) != 4 {
		t.Errorf("Expected 4 elements, got %d", len(array))
	}
}

// TestPersistedState tests maintaining state between executions using upvalues
func TestPersistedState(t *testing.T) {
	// Create config with a stateful module
	config := NewWorkerConfig()

	// Add a module that has internal state via upvalues
	counterModule := `
		-- Module with internal counter
		local counter = 0
		
		local stateful = {}
		
		function stateful.increment()
			counter = counter + 1
			return counter
		end
		
		function stateful.get_value()
			return counter
		end
		
		function stateful.reset()
			counter = 0
			return counter
		end
		
		return stateful
	`
	config.AddModule("counter", counterModule)

	// Create pool with single worker to ensure same Lua state is used
	pool, err := NewPool(1, config)
	if err != nil {
		t.Fatalf("Failed to create pool: %v", err)
	}
	defer pool.Shutdown()

	// First job: increment counter
	bytecode1 := createTestBytecode(t, `
		local counter = require("counter")
		return counter.increment()
	`)

	result1, err := pool.Submit(bytecode1, nil)
	if err != nil {
		t.Fatalf("Failed to submit first job: %v", err)
	}

	if result1 != 1.0 {
		t.Errorf("Expected counter value 1, got %v", result1)
	}

	// Second job: increment counter again
	bytecode2 := createTestBytecode(t, `
		local counter = require("counter")
		return counter.increment()
	`)

	result2, err := pool.Submit(bytecode2, nil)
	if err != nil {
		t.Fatalf("Failed to submit second job: %v", err)
	}

	if result2 != 2.0 {
		t.Errorf("Expected counter value 2, got %v", result2)
	}

	// Third job: get current value without incrementing
	bytecode3 := createTestBytecode(t, `
		local counter = require("counter")
		return counter.get_value()
	`)

	result3, err := pool.Submit(bytecode3, nil)
	if err != nil {
		t.Fatalf("Failed to submit third job: %v", err)
	}

	if result3 != 2.0 {
		t.Errorf("Expected counter value 2, got %v", result3)
	}
}

// TestComplexConfiguration tests a more complex setup with multiple features
func TestComplexConfiguration(t *testing.T) {
	// Create config with multiple features
	config := NewWorkerConfig()

	// Add Go functions
	config.AddFunction("add", func(s *luajit.State) int {
		a := s.ToNumber(1)
		b := s.ToNumber(2)
		s.PushNumber(a + b)
		return 1
	})

	// Add a module
	mathModule := `
		local math_ext = {}
		
		function math_ext.square(x)
			return x * x
		end
		
		function math_ext.cube(x)
			return x * x * x
		end
		
		return math_ext
	`
	config.AddModule("math_ext", mathModule)

	// Set custom package path
	config.SetPackagePath("./?.lua;./test/?.lua")

	// Set custom init function
	config.SetCustomInit(func(s *luajit.State) error {
		return s.DoString(`
			-- Set a global helper
			function capitalize(s)
				return string.upper(string.sub(s, 1, 1)) .. string.sub(s, 2)
			end
		`)
	})

	// Create pool with config
	pool, err := NewPool(2, config)
	if err != nil {
		t.Fatalf("Failed to create pool: %v", err)
	}
	defer pool.Shutdown()

	// Test using all the features together
	bytecode := createTestBytecode(t, `
		local math_ext = require("math_ext")
		
		local sum = add(10, 5)
		local squared = math_ext.square(sum)
		local name = capitalize("test")
		
		-- Check package.path was set correctly
		local has_custom_path = string.find(package.path, "./test/?.lua") ~= nil
		
		return {
			sum = sum,
			squared = squared,
			name = name,
			has_custom_path = has_custom_path
		}
	`)

	result, err := pool.Submit(bytecode, nil)
	if err != nil {
		t.Fatalf("Failed to submit job: %v", err)
	}

	// Check results
	resultMap, ok := result.(map[string]any)
	if !ok {
		t.Fatalf("Expected map result, got %T", result)
	}

	if resultMap["sum"] != 15.0 {
		t.Errorf("Expected sum=15, got %v", resultMap["sum"])
	}

	if resultMap["squared"] != 225.0 {
		t.Errorf("Expected squared=225, got %v", resultMap["squared"])
	}

	if resultMap["name"] != "Test" {
		t.Errorf("Expected name=Test, got %v", resultMap["name"])
	}

	if resultMap["has_custom_path"] != true {
		t.Errorf("Expected has_custom_path=true, got %v", resultMap["has_custom_path"])
	}
}

// TestWorkerConfigInitFail tests error handling during worker initialization
func TestWorkerConfigInitFail(t *testing.T) {
	// Create config with invalid Lua code that will fail
	config := NewWorkerConfig()

	// Set custom init function that will definitely fail
	config.SetCustomInit(func(s *luajit.State) error {
		return s.DoString(`
			-- This has a syntax error
			function without_end(
		`)
	})

	// Create pool - should still work even though module is invalid
	pool, err := NewPool(1, config)
	if err != nil {
		t.Fatalf("Pool creation should succeed despite bad module: %v", err)
	}
	defer pool.Shutdown()

	// Worker count should be 0 as initialization failed
	if pool.ActiveWorkers() != 0 {
		t.Errorf("Expected 0 active workers, got %d", pool.ActiveWorkers())
	}
}
